from notter.controller import NoteController
from notter.notter import Notter


class NotterContext:
    def __init__(self, notter: Notter, controller: NoteController):
        self.notter = notter
        self.controller = controller



from datetime import datetime
from pathlib import Path
from typing import List

import notter.constants as ncons
from notter.exceptions import NotterException
from notter.index import NoteIndex
from notter.model import Comment, Content, Note, NoteWithContent
from notter.notter import Notter


class BaseRepository:
    def __init__(self, notter: Notter) -> None:
        raise NotImplementedError

    def create(self, note_with_content: NoteWithContent) -> None:
        raise NotImplementedError

    def read(self, filepath: str, line: int) -> NoteWithContent:
        raise NotImplementedError

    def update(self, filepath: str, line: int, note_with_content: NoteWithContent) -> None:
        raise NotImplementedError

    def delete(self, filepath: str, line: int) -> None:
        raise NotImplementedError


class JsonFileRepository(BaseRepository):
    def __init__(self, notter: Notter) -> None:
        self.notter = notter

        idx_initialized = self.notter.get_config(ncons.IDX_INITIALIZED_FLAG)
        notter_path = Path(self.notter.get_config(ncons.PATH))
        note_idx_path = str(notter_path / ncons.NOTES_INDEX_FILENAME)

        self.note_index: NoteIndex = NoteIndex(note_idx_path)

        if not idx_initialized:
            self.note_index.init_index(note_idx_path)
            self.notter.set_config(ncons.NOTES_INDEX_PATH, note_idx_path)
            self.notter.set_config(ncons.IDX_INITIALIZED_FLAG, True)
        else:
            self.note_index.load_index(note_idx_path)

    def create(self, note_with_content: NoteWithContent) -> None:
        # TODO: Make these two operations atomic
        # Update the index
        self.note_index.store(note_with_content.note)
        # Write note content to a file
        with open(self._get_note_content_path(note_with_content.note.id), "w") as note_file:
            note_file.write(note_with_content.content.text)

    def read(self, filepath: str, line: int) -> NoteWithContent:
        entry = self.note_index.fetch(filepath, str(line))
        note = Note(**entry.__dict__)
        with open(self._get_note_content_path(note.id), "r") as note_file:
            text = note_file.read()

        content = Content(text)
        return NoteWithContent(note, content)

    def update(self, filepath: str, line: int, note_with_content: NoteWithContent) -> None:
        existing_note: NoteWithContent = self.read(filepath, line)
        existing_note.note.username = note_with_content.note.username
        existing_note.note.email = note_with_content.note.email
        existing_note.note.filepath = note_with_content.note.filepath
        existing_note.note.line = note_with_content.note.line
        existing_note.note.type = note_with_content.note.type
        existing_note.note.updated_at = datetime.now().isoformat()
        existing_note.content = note_with_content.content

        # TODO: Make these two operations atomic
        # Update the index
        self.note_index.store(existing_note.note, True)
        # Write note content to a file
        with open(self._get_note_content_path(existing_note.note.id), "w") as note_file:
            note_file.write(existing_note.content.text)

    def delete(self, filepath: str, line: int) -> None:
        # TODO: Make these two operations atomic
        # Update the index
        entry = self.note_index.clean(filepath, str(line))
        note = Note(**entry)
        note_path = self._get_note_content_path(note.id)
        # Remove note content
        Path(note_path).unlink()

    def prune(self, comments: List[Comment]) -> List[str]:
        comments_set = set()
        for comment in comments:
            comments_set.add(f"{comment.filepath}:{comment.line}")

        entry_set = self.note_index.summarize()
        items_to_prune = list(entry_set.difference(comments_set))
        for item in items_to_prune:
            filepath, line = item.split(":")
            try:
                self.delete(filepath, int(line))
            except NotterException:
                pass

        return items_to_prune

    def _get_note_content_path(self, note_id: str) -> str:
        notes_folder = Path(self.notter.get_config(ncons.NOTES_PATH))
        return str(notes_folder / note_id)



import json
import os
import subprocess
from importlib.metadata import version as pkg_version
from pathlib import Path
from typing import Optional, Tuple

import click
from click import Context, pass_context

import notter.constants as ncons
from notter.context import NotterContext
from notter.controller import NoteController
from notter.exceptions import NotterException
from notter.model import NoteType
from notter.notter import Notter

SRC_PATH_VAR = "SRC_PATH"
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def fetch_git_config() -> Tuple[str, str]:
    click.echo("Binding your Git user to Notter")
    username_process = subprocess.run(["git", "config", "user.name"], capture_output=True)
    email_process = subprocess.run(["git", "config", "user.email"], capture_output=True)

    if username_process.returncode != 0 or email_process.returncode != 0:
        click.secho(
            "Could not fetch your Git username/email, please make sure that `git config` command works.",
            fg="red",
        )
        quit()

    username = username_process.stdout.decode("utf-8").strip("\n")
    email = email_process.stdout.decode("utf-8").strip("\n")
    if not username or not email:
        click.secho("Git username/email not configured properly.", fg="red")
        quit()

    click.secho(f"Using Git username: {username}", fg="yellow")
    return username, email


@click.group(invoke_without_command=True)
@click.option("--init", is_flag=True, help="Initialize Notter tool.")
@click.option("--version", is_flag=True, help="Print Notter version.")
@click.pass_context
def cli(ctx: Context, init: bool, version: bool) -> None:
    src_path = os.getenv(SRC_PATH_VAR)
    if not src_path:
        click.secho(
            f"Could not find the source folder. Please export your source \
                folder as the environment variable: `{SRC_PATH_VAR}`",
            fg="red",
        )
        quit()

    # Initialize Notter instance
    notter = Notter()

    if version:
        click.echo(f'{pkg_version("notter")}')
        return

    if not init:
        notter.load(src_path)
    else:
        # TODO: Do not initialize if the Notter instance is already there
        username, email = fetch_git_config()
        click.secho(f"Initializing Notter with `{src_path}` as source folder.", fg="yellow")
        notter.configure(Path(src_path).resolve())
        notter.set_config(ncons.USERNAME, username)
        notter.set_config(ncons.EMAIL, email)

    controller = NoteController(notter)
    # Add a custom object to context so they are available for other commands
    if ctx.obj is None:
        ctx.obj = NotterContext(notter, controller)


@cli.command()
@click.option("--get", help="get value of a config")
@click.option("--set", "setc", nargs=2, type=str, help="set a config with a key-value pair")
@click.option("--all", "allc", is_flag=True, help="show all the config")
@pass_context
def config(ctx: Context, get: Optional[str], setc: Optional[Tuple[str, str]], allc: bool) -> None:
    """Configure Notter configuration."""
    notter = ctx.obj.notter
    if get:
        value = notter.get_config(get)
        click.echo(value)
    elif setc:
        key, value = setc
        notter.set_config(key, value)
    elif allc:
        click.echo(notter.repr_config())


@cli.command()
@click.pass_context
def destroy(ctx: Context) -> None:
    """Deletes configured Notter instance"""
    notter = ctx.obj.notter
    notter_path = notter.get_config(ncons.PATH)
    click.confirm(f"Do you really want to delete the Notter instance initialized at `{notter_path}` ?", abort=True)
    notter.destroy()
    ctx.obj = None


@cli.command()
@click.argument("filepath", type=str)
@click.argument("line", type=int)
@click.argument("text", type=str)
@click.argument("type", type=NoteType)
@pass_context
def create(ctx: Context, filepath: str, line: int, text: str, type: NoteType) -> None:
    try:
        ctx.obj.controller.create(filepath, line, text, type)
        click.secho("Note created", fg="green")
    except NotterException as exc:
        click.secho(exc.message, fg="red")


@cli.command()
@click.argument("filepath", type=str)
@click.argument("line", type=int)
@pass_context
def read(ctx: Context, filepath: str, line: int) -> None:
    try:
        note = ctx.obj.controller.read(filepath, line)
        click.echo(note)
    except NotterException as exc:
        click.secho(exc.message, fg="red")


@cli.command()
@click.argument("filepath", type=str)
@click.argument("line", type=int)
@click.argument("text", type=str)
@click.argument("type", type=NoteType)
@pass_context
def update(ctx: Context, filepath: str, line: int, text: str, type: NoteType) -> None:
    try:
        ctx.obj.controller.update(filepath, line, text, type)
        click.secho("Note updated", fg="green")
    except NotterException as exc:
        click.secho(exc.message, fg="red")


@cli.command()
@click.argument("filepath", type=str)
@click.argument("line", type=int)
@pass_context
def delete(ctx: Context, filepath: str, line: int) -> None:
    try:
        ctx.obj.controller.delete(filepath, line)
        click.secho("Note deleted", fg="green")
    except NotterException as exc:
        click.secho(exc.message, fg="red")


@cli.command()
@pass_context
def discover(ctx: Context) -> None:
    try:
        # TODO: Retrieve tags from user
        tags = ["TODO", "FIXME"]
        comments = ctx.obj.controller.discover(tags)
        click.echo(json.dumps(comments, default=lambda o: o.__dict__))
    except NotterException as exc:
        click.secho(exc.message, fg="red")



import glob
import io
import tokenize
from pathlib import Path
from typing import List

import notter.constants as ncons
from notter.model import Comment, NoteType
from notter.notter import Notter


class BaseExplorer:
    def __init__(self, notter: Notter):
        raise NotImplementedError

    def discover(self, tags: List[str]) -> List[Comment]:
        raise NotImplementedError


class LexicalExplorer(BaseExplorer):
    def __init__(self, notter: Notter) -> None:
        self.source_path = notter.get_config(ncons.SRC_PATH)

    def discover(self, tags: List[str]) -> List[Comment]:
        comments = []
        tags = [tag.lower() for tag in tags]
        glob_pattern = str(Path(self.source_path) / "**/*.py")

        for file in glob.glob(glob_pattern, recursive=True):
            comments_in_file = LexicalExplorer._discover_comments_in_file(file, tags)
            comments.extend(comments_in_file)

        return comments

    @staticmethod
    def _discover_comments_in_file(filepath: str, tags: List[str]) -> List[Comment]:
        comments: List[Comment] = []

        with open(filepath, "r", encoding="utf-8") as file:
            file_content = file.read()

        tokens = tokenize.tokenize(io.BytesIO(file_content.encode()).readline)
        for token_type, token_content, token_start, _, _ in tokens:
            if token_type == tokenize.COMMENT:
                token_content = token_content[1:]

                is_todo = False
                for tag in tags:
                    if tag in token_content.lower():
                        is_todo = True

                if comments and token_start[0] == (comments[-1].line + 1) and not is_todo:
                    comments[-1].text += f"\n{token_content}"
                    comments[-1].multiline = True
                else:
                    note_type = NoteType.TODO if is_todo else NoteType.NOTE
                    comments.append(Comment(filepath, token_content, token_start[0], note_type))

        return comments



from typing import List

import notter.constants as ncons
from notter.exceptions import NoteAlreadyExists
from notter.explorer import LexicalExplorer
from notter.model import Comment, Content, Note, NoteType, NoteWithContent
from notter.notter import Notter
from notter.repository import JsonFileRepository


class NoteController:
    def __init__(self, notter: Notter):
        self.notter = notter
        self.repository = JsonFileRepository(self.notter)
        self.explorer = LexicalExplorer(self.notter)

    def _create_note_with_content(self, filepath: str, line: int, text: str, type: NoteType) -> NoteWithContent:
        username = self.notter.get_config(ncons.USERNAME)
        email = self.notter.get_config(ncons.EMAIL)
        # TODO: Add input validation, max number of characters for each field
        note = Note(username, email, filepath, line, type)
        content = Content(text)
        return NoteWithContent(note, content)

    def create(self, filepath: str, line: int, text: str, type: NoteType = NoteType.NOTE) -> None:
        note_with_content = self._create_note_with_content(filepath, line, text, type)
        self.repository.create(note_with_content)

    def read(self, filepath: str, line: int) -> NoteWithContent:
        return self.repository.read(filepath, line)

    def update(self, filepath: str, line: int, text: str, type: NoteType = NoteType.NOTE) -> None:
        note_with_content = self._create_note_with_content(filepath, line, text, type)
        self.repository.update(filepath, line, note_with_content)

    def delete(self, filepath: str, line: int) -> None:
        self.repository.delete(filepath, line)

    def discover(self, tags: List[str], save_as_notter_notes: bool = True) -> List[Comment]:
        comments: List[Comment] = self.explorer.discover(tags)

        if save_as_notter_notes:
            for comment in comments:
                try:
                    self.create(comment.filepath, comment.line, comment.text, comment.type)
                except NoteAlreadyExists:
                    continue

        _ = self.repository.prune(comments)
        return comments



import json
from typing import Dict, Optional, Set

from notter.exceptions import NoteAlreadyExists, NoteNotFound
from notter.model import Note
from notter.utils import CustomEncoder, persist_index_after

Line = str
FilePath = str
NoteDict = Dict[FilePath, Dict[Line, Note]]


class NoteIndex:
    def __init__(self, index_path: str) -> None:
        self.idx: NoteDict = {}
        self.idx_path = index_path

    def init_index(self, index_path: str) -> None:
        with open(index_path, "w+") as idx_file:
            json.dump(self.idx, idx_file, cls=CustomEncoder, indent=4)

    def load_index(self, index_path: str) -> None:
        with open(index_path, "r") as file:
            self.idx = json.loads(file.read())

    def seek(self, filepath: str, line: str) -> bool:
        file_entry = self.idx.get(filepath)
        if not file_entry:
            return False

        note_entry = file_entry.get(line)
        if not note_entry:
            return False

        return True

    def fetch(self, filepath: str, line: str) -> Optional[Note]:
        note_exists = self.seek(filepath, line)
        if not note_exists:
            raise NoteNotFound

        return self.idx[filepath][line]

    def summarize(self) -> Set[str]:
        entry_set: Set[str] = set()
        for filepath, note_dict in self.idx.items():
            for line in note_dict.keys():
                entry_set.add(f"{filepath}:{line}")

        return entry_set

    @persist_index_after
    def store(self, note: Note, update: bool = False) -> None:
        note_exists = self.seek(note.filepath, str(note.line))
        if not update and note_exists:
            raise NoteAlreadyExists

        self.idx[note.filepath] = self.idx.get(note.filepath, {})
        self.idx[note.filepath][str(note.line)] = note

    @persist_index_after
    def clean(self, filepath: str, line: str) -> Optional[Note]:
        note = self.fetch(filepath, line)
        file_entry = self.idx[filepath]
        file_entry.pop(line)

        # Get rid of file entry if it has no keys inside
        if not file_entry:
            self.idx.pop(filepath)

        return note



PATH = "path"
SRC_PATH = "src_path"
CONFIG_PATH = "config_path"
USERNAME = "username"
EMAIL = "email"

NOTES_PATH = "notes_path"
NOTES_INDEX_PATH = "notes_index_path"
NOTES_INDEX_FILENAME = "idx_notes.json"

CONFIG_FILENAME = "config.json"
NOTES_DIRNAME = "notes"

INITIALIZED_FLAG = "initialized"
IDX_INITIALIZED_FLAG = "idx_initialized"



from pathlib import Path
from shutil import rmtree
from typing import Any, Dict

import click

import notter.constants as ncons
from notter.utils import load_config, persist_config_after


class Notter:
    def __init__(self) -> None:
        self.config: Dict[str, Any] = {}
        self.config[ncons.INITIALIZED_FLAG] = False
        self.config[ncons.IDX_INITIALIZED_FLAG] = False

    @persist_config_after
    def configure(self, src_folder: Path) -> None:
        self.path = src_folder.parent / ".notter"
        self.config[ncons.CONFIG_PATH] = str(self.path / ncons.CONFIG_FILENAME)
        self.config[ncons.SRC_PATH] = str(src_folder)
        self.config[ncons.PATH] = str(self.path)
        self.config[ncons.NOTES_PATH] = str(self.path / ncons.NOTES_DIRNAME)
        self.init_notter_folders()

    def init_notter_folders(self) -> None:
        if self.get_config(ncons.INITIALIZED_FLAG):
            click.secho(f"Notter folders found at location: {self.path}", fg="red")
        else:
            click.secho(f"Creating Notter folders at location: {self.path}", fg="yellow")
            Path(self.get_config(ncons.PATH)).mkdir(parents=True, exist_ok=True)
            Path(self.get_config(ncons.NOTES_PATH)).mkdir(parents=True, exist_ok=True)
            self.set_config(ncons.INITIALIZED_FLAG, True)

    def load(self, src_folder: str) -> None:
        src_path = Path(src_folder).resolve()
        parent_path = src_path.parent
        notter_path = parent_path / ".notter"
        notter_config_file = str(notter_path / ncons.CONFIG_FILENAME)
        loaded_config = load_config(notter_config_file)
        if not loaded_config:
            click.secho("Could not load Notter configuration", fg="red")
            quit()

        self.config = loaded_config

    def destroy(self) -> None:
        if not self.get_config(ncons.INITIALIZED_FLAG):
            click.secho("No Notter instance is found, nothing to delete", fg="red")
        else:
            # TODO: Make sure it is not a critical path
            notter_path = self.get_config(ncons.PATH)
            rmtree(notter_path)
            click.secho("Notter instance deleted", fg="green")

    def repr_config(self) -> str:
        repr = "Notter config:\n==============\n"
        configs_str = "\n".join([f"{key}: {value}" for key, value in self.config.items()])
        return repr + configs_str

    def get_config(self, key: str) -> Any:
        value = self.config.get(key, None)
        if value is None:
            click.secho(f"Config `{key}` is not set", fg="red")
            return None

        return value

    @persist_config_after
    def set_config(self, key: str, value: Any) -> None:
        self.config[key] = value
        click.secho(f"Set config `{key}` to `{value}`", fg="yellow")



from pathlib import Path
from unittest.mock import patch
from click.testing import CliRunner
from notter.cli import cli
from notter.notter import Notter

from importlib.metadata import version as pkg_version


class TestCLI:
    def test_cli_version(self):
        runner = CliRunner(env={"SRC_PATH": "."})
        result = runner.invoke(cli, ["--version"])
        assert result.exit_code == 0
        assert result.output.strip("\n") == pkg_version("notter")



import uuid
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional


class NoteType(str, Enum):
    NOTE = "NOTE"
    TODO = "TODO"


# TODO: Add input validation, max number of characters for each field
@dataclass
class Note:
    username: str
    email: str
    filepath: str
    line: int
    type: NoteType = NoteType.NOTE
    id: str = str(uuid.uuid4())
    created_at: Optional[str] = datetime.now().isoformat()
    updated_at: Optional[str] = datetime.now().isoformat()


@dataclass
class Content:
    text: str


@dataclass
class NoteWithContent:
    note: Note
    content: Content

    def __str__(self) -> str:
        return f"{self.content.text} - {self.note.username} / {self.note.updated_at}"


@dataclass
class Comment:
    filepath: str
    text: str
    line: int
    type: NoteType
    multiline: bool = False

    def __str__(self) -> str:
        return f"{self.filepath}:{self.line} - {self.text}"



from setuptools import setup

setup(
    name="notter",
    version="0.1.0",
    description="A simple note taking tool for software projects.",
    author="Taylan Dogan",
    author_email="taylandogan.nl@gmail.com",
    py_modules=["notter"],
    include_package_data=True,
    install_requires=["click", "pygments"],
    extras_require={
        "dev": ["black", "flake8", "isort", "mypy", "coverage", "pytest"],
    },
    entry_points={
        "console_scripts": ["notter=notter.cli:cli"],
    },
)



import json
from typing import Any, Dict, Optional

import click

import notter.constants as ncons


class CustomEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__


def load_config(config_file: str) -> Optional[Dict[str, Any]]:
    config_data = None
    try:
        with open(config_file, "r") as file:
            config_data = json.loads(file.read())
    except FileNotFoundError:
        click.secho(f"Could not load Notter config from: {config_file}", fg="red")

    return config_data


def persist_config_after(function):
    def wrapper(self, *args, **kwargs):
        retval = function(self, *args, **kwargs)

        config_file = self.get_config(ncons.CONFIG_PATH)
        with open(config_file, "w+") as file:
            json.dump(self.config, file, indent=4)

        click.secho("Notter config updated", fg="green")
        return retval

    return wrapper


def persist_index_after(function):
    def wrapper(self, *args, **kwargs):
        retval = function(self, *args, **kwargs)

        with open(self.idx_path, "w") as idx_file:
            json.dump(self.idx, idx_file, cls=CustomEncoder, indent=4)

        click.secho("Notter index updated", fg="green")
        return retval

    return wrapper



class NotterException(Exception):
    message = "Something went wrong, please inform the maintainer"


class NoteNotFound(NotterException):
    message = "Could not find that note"


class NoteAlreadyExists(NotterException):
    message = "That node already exists"
